#!/bin/python
# rubika clietn self channel methods

from requests       import get, post
from pathlib        import Path
from random         import choice, randint, sample
from re             import findall, search, sub, compile, escape
from time           import sleep, gmtime, localtime
from datetime       import datetime
from json           import load, loads, dumps
from .crypto        import Encryption
from .connection    import GetData, Urls
from .UserMethods   import UserMethods
from .clients       import *
from .exceptions    import *
from .extensions    import *
from .sessions      import RubikaClient as Client
import typing


class ChannelMethods:


    __extend_class__ = 'ChannelMethods'


    def __enter__(self):
        return (self)

    def __exit__(self, *args,
                 **kwargs) -> None:
        pass


    set_channel_link        =   lambda self, chat_id: GetData.api(version=self.api_version or '5', method='setChannelLink', auth=self.auth, data={'channel_guid': chat_id}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    check_channel_username  =   lambda self, username: GetData.api(version=self.api_version or '5', method='checkChannelUsername', auth=self.auth, data={'username': username}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)
    delete_channel          =   lambda self, chat_id: GetData.api(version=self.api_version or '5', method='removeChannel', auth=self.auth, data={'channel_guid': chat_id}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)


    def add_channel_members(
    
        self        :   'Client',
        user_ids    :   (list),
        chat_id     :   (str)
        ) -> typing.Union[str, dict]:

            '''

            self.add_channel_members(['user guid', ]. chat_id='channel guid')
            
            PARAMETERS:
                1- user_ids a list user guids
                3- chat_id is channel guid
            '''

            return (GetData.api(version = self.api_version or '5', method = 'addChannelMembers', auth = self.auth,
                                data = {'channel_guid': chat_id, 'member_guids': user_ids}, mode = self.city,
                                platform = self.platform or 'web', proxy = self.proxy))


    def get_channel_all_members(
        self            :   ('Client'),
        chat_id         :   (str)   =   (None),
        search_text     :   (str)   =   (None),
        start_id        :   (str)   =   (None)
        ) -> typing.Union[str, dict]:
            
            '''
            self.get_channel_all_members(...)
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'getChannelAllMembers',
                    auth        =   self.auth,
                    data        =   {
                        'channel_guid'  :   chat_id,
                        'search_text'   :   search_text,
                        'start_id'      :   start_id
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def set_channel_action(
        self    :   Client,
        chat_id :   str,
        action  :   str =   'Join'
        ) -> typing.Union[str, dict]:

            '''
            this method for join and leave channels
            
            USE:
                self.set_action_chat('chat-guid', 'Pin')
            PARAMS:
                1- self is a self object
                2- chat_id is chat guid
                3- action is a action type. actions: 'Join', 'Leave'
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'joinChannelAction',
                    data        =   {
                        'channel_guid'   :   chat_id,
                        'action'        :   action
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def add_channel(
        self            : Client,
        title           : str,
        description     : str,
        channel_type    : str   =   'Public'
        ) -> typing.Union[str, dict]:
            
            '''
            this method for create channel
            
            USE:
                self.add_channel('title', 'bio')
            PARAMS:
                1- self is a self object
                2- title is channel name
                3- description is channel bio
                4- channel_type is a action type for channel. types: 'Public', 'Private'
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'addChannel',
                    data        =   {
                        'title'         :   title,
                        'description'   :   description,
                        'channel_type'  :   channel_type
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def edit_channnel_info(
        self    :   Client,
        chat_id :   str,
        **kwargs:   dict or str
        ) -> typing.Union[str, dict]:

            '''
            USE:
                self.edit_channnel_info('chat-guid', title='name')
            PARAMS:
                sign_messages, title, description
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'editChannelInfo',
                    data        =   {
                        'updated_parameters'        :   list(kwargs.keys())
                        }.update(kwargs),
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def update_channel_username(
        self    :   Client,
        chat_id :   str,
        username:   str
        ) -> typing.Union[str, dict]:
            
            '''
            `self.update_channel_username('channel-guid', 'username')`
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'updateChannelUsername',
                    data        =   {
                        'channel_guid'  :   chat_id,
                        'username'      :   username
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def set_channel_admin(self: Client, chat_id: str,
                          member_guid: str, access_list: list = None,
                          action: str = 'SetAdmin') -> typing.Union[str, dict]:
        
        '''
        ## this method is for admin and unadmin a member chat
        
        
        USE:
            `self.set_channel_admin('group-guid', 'user-guid', [Accesses.admin.SendMessages])`
        
        PARAMS:
            1- `self` a is self object
            2- `chat_id` is group guid
            3- `member_guid` is a user guid
            4- `access_list` is for access user in group: from rub import accesses.admin or ["ChangeInfo", "ViewMembers", "ViewAdmins", "PinMessages", "SendMessages", "EditAllMessages", "DeleteGlobalAllMessages", "AddMember", "SetJoinLink"]
            5- `action` is action type: 'UnsetAdmin', 'SetAdmin'
        '''
        
        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'setChannelAdmin',
                    data        =   {
                        'channel_guid'  :   chat_id,
                        'member_guid'   :   member_guid,
                        'action'        :   action,
                        'access_list'   :   access_list
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def channel_preview_by_join_link(self: Client, link: str) -> typing.Union[str, dict]:
        
        '''
        get channel info from link
        '''
        
        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'channelPreviewByJoinLink',
                    data        =   {
                        'hash_link' :       findall(r'([A-Z0-9]+)$', link)[0]
                      },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def join_channel_by_link(self: Client, link: str) -> typing.Union[str, dict]:

        '''
        this method for join channel with link

        `self.join_channel_by_link('https://rubika.ir/joinc/...')`
        '''

        return (GetData.api(version = self.api_version or '4', auth = self.auth,
                            method = 'joinChannelByLink', data = {'hash_link': findall(r'([A-Z0-9]+)$', link)[0]}, mode = self.city,
                            platform = self.platform or 'android', proxy = self.proxy))


__all__ = ['ChannelMethods']


#!/bin/python

from json           import  dumps, loads, dump, load, JSONDecodeError
from random         import  choice, randint
from requests       import  get, post, session
from .exceptions    import  *
from .crypto        import  Encryption
from .storage       import  SQLiteSession
from .clients       import  *
from .models        import Attrs
import typing, re


class Errors(object):
    
    @classmethod
    def MadeError(cls, status: str,
                  det: str) -> typing.Union[int, bool]:

        if status.upper() != 'OK':
            if 'NOT_REGISTERED' in det.upper():
                raise NotREGISTERED('Session key not found, please find your key in web.rubika.ir and try again.')
            elif 'INVALID_INPUT' in det.upper():
                raise InvalidInput('Your inserts is invalid, please try again.')
            elif 'TOO_REQUESTS' in det.upper():
                raise TooREQUESTS('Method has been limited, please try again later.')
            elif 'INVALID_AUTH' in det.upper():
                raise InvalidAUTH('Server error or please check method arguments and try again.')
            else:
                raise APIError('Request is invalid: (%s)' % status)
        else:
            return 1


class ClientConnectorError(object):
    
    @classmethod
    def __init__(cls, **kwargs) -> None:
        
        '''
        class StorageError(OSError): pass
        
        error = ClientConnectorError(name='[StorageError] dont access to your device.', error=StorageError)
        
        print(error.returns)
        error.warnings
        error.raises
        '''
        
        cls.set_error, cls.name = kwargs.get('error'), kwargs.get('name')

    @property
    def raises(cls) -> (Exception):
        raise cls.set_error(f'this error for {cls.name}')
    
    @property
    def returns(cls) -> (str):
        return 'this error for %s please fixed and try again.' %  cls.name

    @property
    def warnings(cls) -> (str):
        __import__('warnings').warn(cls.name)


class Urls(str):

    def get_url() -> (str):

        '''
        getting a rubika client url from getdcmess or memory session
        '''
        
        for i in range(3):
            try:
                return choice(list(loads(__import__('urllib').request.urlopen('https://getdcmess.iranlms.ir/').read().decode('utf-8')).get('data').get('API').values()))
            except Exception:
                if i == 2:
                    return Urls.giveUrl()
                else:
                    continue

    giveUrl = lambda mode='mashhad', key=None: SQLiteSession(key).information()[3] if key and 'https://' in SQLiteSession(key).information() else ('https://messengerg2c{}.iranlms.ir/'.format(str('56' if mode.lower() == 'mashhad' else '74' if mode.lower() == 'tehran' else str(randint(3, 74)))))


class MakeDataKeys(object):

    def __str__(self) -> (str):
        '''
        return a string of json database storage
        '''
        
        return dumps(self.result_data, indent=4,
                     ensure_ascii=False, default=str)

    def __getattr__(self, name) -> (list):
        return self.keys_finder(keys=name)

    def __setitem__(self, key, value) -> (None):
        self.result_data[key] = value

    def __getitem__(self, key):
        return self.result_data[key]

    def __lts__(self, update: list) -> (list):
        
        for index, element in enumerate(update):
            if isinstance(element, list):
                update[index] = self.__lts__(update=element)

            elif isinstance(element, dict):
                update[index] = MakeDataKeys(update=element)

            else:
                update[index] = element
        return update

    def __init__(self, update: dict = None) -> (None):
        
        '''
        # this object for get keys from a dictionary
        
        `result.data.chat`
        `result.get('data').get('chat')`
        
        '''
        
        self.result_data = update

    def keys_finder(self, keys: list,
                  result_data: str = None) -> (list):

        if not result_data:
            result_data = self.result_data

        if not isinstance(keys, list):
            keys = [keys]

        if isinstance(result_data, dict):
            for key in keys:
                try:
                    update = result_data[key]
                    if isinstance(update, dict):
                        update = MakeDataKeys(update=update)

                    elif isinstance(update, list):
                        update = self.__lts__(update=update)

                    return update

                except KeyError:
                    pass
            result_data = result_data.values()

        for value in result_data:
            if isinstance(value, (dict, list)):
                try:
                    return self.keys_finder(keys=keys, result_data=value)

                except AttributeError:
                    pass
        else:
            ClientConnectorError(name='data has not (%s) key' % ', '.join(keys), error=KeyError).raises


class Make(object):


    action, type = 'dict', 'result'


    def evolution(message: dict, key: str,
                  mode: bool = False,
                  *args, **kwargs) -> typing.Union[dict, MakeDataKeys, Attrs]:
        
        '''
        decrypt json response with session
        '''
        
        res: dict = {}

        try:
            res: dict = loads(Encryption(key).decrypt(message.get('data_enc')))
        except Exception:
            res = str(Encryption(key).decrypt(message.get('data_enc')))

        if isinstance(res, dict) and Errors.MadeError(res.get('status') or '', res.get('status_det') or ''):
            
            if Make.type.__eq__('message'):
                res: dict = res.get('data')
                res: dict = res.get('messages') or res.get('chats') or res.get('in_chat_members') or res.get('message_updates')
                return [Attrs.create(res, action='message') for data in res]
            
            elif Make.type.__eq__('bot'):
                res: dict = res.get('data')
                
            return MakeDataKeys(res) if Make.action.__contains__('str') else Attrs.create(res, action=Make.type) if Make.action.__eq__('object') else (res)
        
        else:
            return res


    def base_pwa(__method: str = 'getBaseInfo', /,
                 __data: dict = {}, __session: str = None,
                 *args, **kwargs) -> typing.Union[dict, MakeDataKeys, Attrs]:

        '''
        this method has for make pwa platform
        '''

        result: dict = session().post('https://servicesbase.iranlms.ir/', json=({'method': __method, 'api_version': '0', 'data': __data, 'auth': __session, 'client': clients.pwa}), timeout=5).json()
        
        return MakeDataKeys(result) if Make.action.__contains__('str') else Attrs.create(result, action=Make.type) if Make.action.__eq__('object') else (result)


class Connection(dict):

    timeout, headers, api_version, platform, action, proxy, type = 5, None, '5', 'web', 'sync', {'http': 'http://127.0.0.1:9050'}, True
    # TODO: the set all variable to params from init object and dont use on methods func.
    
    @staticmethod
    def postion(
        url     :   (str),
        data    :   (dict),
        proxy   :   (dict),
        auth    :   (str),
        mode    :   (bool) = (False)
        ) -> typing.Dict[str, int]:
        
        '''
        postion has for request http & post data to client
        '''
        
        with session() as (sent):

            for (i) in range(3):
                
                try:
                    if Connection.action == 'async':
                        pass
                    else:
                        return (Make.evolution((sent).post(url, json=(data) if not mode else dumps(data), timeout=Connection.timeout, proxies=proxy).json(), (auth), True if not auth else False))
                
                except JSONDecodeError:
                    ...
                except Exception as e:
                        raise e
                finally:
                    pass

            else:
                raise ServerError('Device can\'t connect to the server and is not response, please checked your network.')


class GetData(Connection):


    url: typing.Union[str, bool] = False


    @staticmethod
    def api(*args, **kwargs) -> (typing.Union[typing.Dict[str, int], Connection]):

        '''
        # CLIENT METHODS
        
        ## EXAMPLE:
        
            - `version` =   '5' or '4',
            - `auth`    =   'key',
            - `tmp`     =   ...
            - `method`  =   'methodName'
            - `data`    =   input,
            - `mode`    =  'mashhad',
            - `platform`=   'rubx' or 'web',
            - `proxy`   =   {'https':'127.0.0.1:9050'} # a dictionary type
        '''

        main: list = []
        
        if (kwargs.get('version') == '5'):
            main.extend(
                [
                    {
                        'api_version': '5',
                        '{}'.format('auth' if kwargs.get('auth') else 'tmp_session') :   kwargs.get('auth') or kwargs.get('tmp'),
                        'data_enc': ({'input': kwargs.get('data'), 'client': clients.web if kwargs.get('platform') == 'web' else clients.rubx, 'method': kwargs.get('method')}) if not kwargs.get('auth') else Encryption(kwargs.get('auth')).encrypt(dumps({'input': kwargs.get('data'), 'client': clients.web if kwargs.get('platform') == 'web' else clients.rubx, 'method': kwargs.get('method')}))
                        }
                    ]
                )
        
        else:
            main.extend(
                [
                    {
                        'api_version'   :   '4',
                        'auth'          :   kwargs.get('auth'),
                        'client'        :   clients.android,
                        'method'        :   kwargs.get('method'),
                        'data_enc'      :   Encryption(kwargs.get('auth')).encrypt(dumps(kwargs.get('data')))
                }
                    ]
                )

        return (Connection.postion(GetData.url or Urls.giveUrl(kwargs.get('mode'), kwargs.get('auth')), main[0], kwargs.get('proxy'), kwargs.get('auth') or kwargs.get('tmp')))


    @classmethod
    def tmp(*args, **kwargs) -> (typing.Union[dict, Connection]):
        
        '''
        tmp for tmp session requests to signup in to rubika
        '''
        
        data: dict = {'api_version': 5, 'tmp_session': kwargs.get('tmp'), 'data_enc': Encryption(kwargs.get('tmp')).encrypt(dumps({'input': kwargs.get('data'), 'method': kwargs.get('method'), 'client': clients.web}))}
        return Connection.postion(GetData.url or Urls.giveUrl(kwargs.get('mode'), None), data, kwargs.get('proxy'), kwargs.get('tmp'))


    def bot(types: typing.Literal['http', 'file'] = 'http',
            *args, **kwargs) -> typing.Union[dict, str]:
        
        '''
        make data from method for posting to api, token client has for usage rubika bot
        '''
        
        # TODO: get api random, rnd: str = str(randint(1, 50))
        
        with session() as sent:
            
            for i in range(3):
                
                try:
                    result: dict = sent.post(url='https://messengerg2b1.iranlms.ir/v3/{}/{}'.format(kwargs.get('token'), kwargs.get('method')), json=dumps(kwargs.get('data')), timeout=Connection.timeout, proxy=Connection.proxy).json() if types.__eq__('http') else loads(sent.post(kwargs.get('url'), files=kwargs.get('data'), verify=False).text)
                except Exception as e:
                    if i == 2:
                        raise BotApiTimeOutError(f'{e}')

        if Errors.MadeError(result.get('status'), 'true'):
            return MakeDataKeys(result) if Connection.type else result


    def pwa(*args, **kwargs) -> Make:
        return Make.base_pwa(*args, **kwargs)




#!/bin/python
# the tools version

import re, json, base64


class Clean(str):

    @classmethod
    def html_cleaner(cls, text: str) -> (str):
        return re.sub(re.compile(r'<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});'), '', text)


class Scanner(object):

    @classmethod
    def check_type(cls, chat_id: str) -> (str):
             
            if len(chat_id) > 2:

                if   (chat_id.lower().startswith('g0')):
                    return 'Group'

                elif (chat_id.lower().startswith('c0')):
                    return 'Channel'

                elif (chat_id.lower().startswith('u0')):
                    return 'User'

                elif (chat_id.lower().startswith('s0')):
                    return 'Service'

                elif (chat_id.lower().startswith('b0')):
                    return 'Bot'
                else                            :
                    raise ValueError(f'CHAT ID \'{chat_id}\' NOT FOUND.')
            else:
                raise ValueError(f'CHAT ID \'{chat_id}\' FALSE.')


class Maker(object):

    @classmethod
    def check_link(cls, link: str, client: object = None,
                   *args, **kwargs) -> (str):

        '''
        this is has a method for get link from a chat id
        
        `self.Make.check_link('https://rubika.ir/*', self)`
        '''

        if link.startswith('@') or re.search(r'rubika\.ir/\w{4,25}', link):

            link: str = link.replace('https://', '').replace('rubika.ir/', '')
            result: dict = client.getObjectByUsername(client, link.replace('@', ''))

            if not isinstance(result, dict):
                result: dict = json.loads(result).get('data')
            else:
                result: dict = result.get('data')

            result: dict = result.get('user') or result.get('channel') or result.get('group')
            
            return result.get('user_guid') or result.get('channel_guid') or result.get('group_guid')


        elif len(link) == 56 or len(link) == 48 and 'joing' in link or 'joinc' in link:

            if 'joinc' in link:

                result = client.groupPreviewByJoinLink(client, link)
                
                if not isinstance(result, dict):
                    return json.loads(result).get('data')['group']['group_guid']
                else:
                    return result.get('data')['group']['group_guid']

            elif 'joing' in link:

                result = client.channelPreviewByJoinLink(client, link)

                if not isinstance(result, dict):
                    return json.loads(result).get('data')['channel']['channel_guid']
                else:
                    return result.get('data')['channel']['channel_guid']


class Login(object):

    @classmethod
    def SignIn(cls, phone: str,
               UserMethods: object = None) -> (str):
        try:
            return UserMethods.sign_in(phone, UserMethods.send_code(phone, input('send type is SMS/Interval : '), password=input('please enter your password : ') if input('insert password y/n : ').lower() == 'y' else None).get('data').get('phone_code_hash'), input('please enter activation code : ')).get('data').get('auth') or '0'
        except Exception:
            from rb import RubikaClient
            return RubikaClient.sign_in(phone, RubikaClient.send_code(phone, input('send type is SMS/Interval : '), password=input('please enter your password : ') if input('insert password y/n : ').lower() == 'y' else None).get('data').get('phone_code_hash'), input('please enter activation code : ')).get('data').get('auth') or '0'


class _Top(object):

    @classmethod
    def __init__(cls, session: str) -> (None):

        '''
        detecting session key to chars echo
        '''

        cls.session = session

    @classmethod
    def __update(cls, key: str, check: bool,
                 is_fake: bool, adds: list, index_list: list) -> dict:
        return {'key': cls.session, 'check': None, 'is_fake': is_fake, 'adds': [*adds], 'index_list': index_list}

    
    @classmethod
    def detecting(cls) -> (dict):

        if isinstance(cls.session, str) and len(cls.session) == 32:
            exits, session, add, index = True, [*cls.session], '', []
            for key, n in zip(cls.session, range(len(cls.session))):
                
                if session.count(key) > 2:
                    exits: bool = True
                    cls.__update(**{'key': cls.session, 'check': None, 'is_fake': exits, 'adds': [*add], 'index_list': index})

                if session.count(key).__eq__(2):
                    add += key
                    index.extend([n])
                    exits: bool = False
            else:
                return cls.__update(**{'key': cls.session, 'check': None, 'is_fake': exits, 'adds': [*add], 'index_list': index})
        else:
            return {'key': cls.session, 'check': None, 'is_fake': True}


class UpToDate(object):

    def __init__(self,
                 version: str, url: str) -> None:

        '''
        upgrade module to last version from github repo
        '''

        self._ver, self._url = version, url

    @property
    def update(self) -> (str):

        if str(get(self._url).text) != self._ver:
            return 'notUpdated'
        else:
            return 'isUpdated'

    @property
    def user(self) -> (None):

        if self.update != 'isUpdated':
            if input('new version rubx now up to date? y/n : ').upper() == 'Y':
                
                os.system('pip install rubx --upgrade')
                
                if platform.system() == 'Windows':
                    os.system('cls')
                else:
                    os.system('clear')


class Thumbnail(object):

    @classmethod
    def __init__(cls, image: bytes) -> None:

        '''
        getting thumbnail from file bytes & make to base64
        '''

        cls.image = image
        if isinstance(cls.image, str):
            cls.image = open(image, 'rb').read()

    @property
    def to_base64(cls) -> str:

        if cls.image:
            return base64.b64encode(cls.image).decode('utf-8')


#!/bin/python

import pydantic, typing

'''
from rb.responses import Self
from rb import RubikaClient

with RubikaClient(...) as client:
    print(client == Self.Text(chat_id=..., text='Hello'))
'''

# Self Message Text Parser
class Self(object):
    
    '''
    from rb.responses import *
    '''
    
    class Text(pydantic.BaseModel):
        chat_id: str
        text: str
        reply_to_message_id: str = None

    class Forward(pydantic.BaseModel):
        from_object_guid: str
        message_ids: typing.Optional[list]
        to_object_guid: str

    class Delete(pydantic.BaseModel):
        chat_id: str
        message_ids: typing.Optional[list]
    
    class Edit(pydantic.BaseModel):
        chat_id: str
        new_text: str

    __all__ = ['Text', 'Forward',
               'Delete', 'Edit']

'''
from rb import RubikaClient
from rb.responses.Bot import Text

with RubikaClient(...) as client:
    print(`client == Text(chat_id=..., text=...))`
'''

# Bot Message Text Parser
class Bot(object):
    
    '''
    from rb.responses.Bot import *
    '''
    
    class Text(pydantic.BaseModel):
        chat_id: str
        text: str
        reply_to_message_id: str = None

    class Forward(pydantic.BaseModel):
        from_chat_id: str
        message_id: typing.Optional[str]
        to_chat_guid: str

    class Delete(pydantic.BaseModel):
        chat_id: str
        message_id: typing.Optional[str]
    
    class Edit(pydantic.BaseModel):
        chat_id: str
        text: str
        message_id: str

    __all__ = ['Text', 'Forward',
               'Delete', 'Edit']

# TODO: add all methods


#!/bin/python

class Filters(str):
    group, channel, user, chat, author = 'group_guid', 'channel_guid', 'user_guid', 'object_guid', 'author_object_guid'

class Performers(str):
    chats_updates, messages_updates, hand_shake = 'ChatsUpdates', 'MessagesUpdates', 'HandShake'

class AccessList(object):
    
    class Admin(str):
        (
            PinMessages,
            setAdmin,
            ChangeInfo,
            BanMember,
            SetJoinLink,
            SetMemberAccess,
            DeleteGlobalAllMessages
            ) = (
                'PinMessages',
                'setAdmin',
                'ChangeInfo',
                'BanMember',
                'SetJoinLink',
                'SetMemberAccess',
                'DeleteGlobalAllMessages'
                )

    class User(str):
        (
            ViewMembers,
            ViewAdmins,
            SendMessages,
            AddMember
            ) = (
                'ViewMembers',
                'ViewAdmins',
                'SendMessages',
                'AddMember'
                )


__all__ = ['Filters', 'Performers', 'AccessList']


#!/bin/python
# rubika client group self methods

from requests       import get, post
from pathlib        import Path
from random         import choice, randint, sample
from re             import findall, search, sub, compile, escape
from time           import sleep, gmtime, localtime
from datetime       import datetime
from json           import load, loads, dumps
from .crypto        import Encryption
from .connection    import GetData, Urls
from .UserMethods   import UserMethods
from .clients       import *
from .exceptions        import *
from .extensions    import *
from .sessions      import RubikaClient as Client
import typing


class GroupMethods:


    __extend_class__ = 'GroupMethods'


    def __enter__(self):
        return (self)

    def __exit__(
        self, *args,
            **kwargs) -> (None):
        pass


    get_display_as_in_group_voice_chat = lambda self, chat_id: GetData.api(version=self.api_version or '5', method='getDisplayAsInGroupVoiceChat', auth=self.auth, data={'chat_guid': chat_id}, mode=self.city, platform=self.platform or 'web', proxy=self.proxy)


    def __link_scanner(self, link: str) -> str:
        
        '''
        this is has a method for check hash link from a private link
        '''

        if not link.__contains__('/') and not search(r'[A-Z0-9]+', link):
            if len(link).__lt__(32):
                raise ValueError('\'link\' parameter is invalid')
            else:
                return findall(r'([A-Z0-9]+)$', link.upper())[0]

        else:
            return findall(r'([A-Z0-9]+)$', link)[0]


    def ban_group_member(
        self            :   ('Client'),
        user_id         :   (str)   =   (None),
        member_username :   (str)   =   (None),
        chat_id         :   (str)   =   (None),
        link            :   (str)   =   (None),
        action          :   (str)   =   ('Set'),
        ) -> typing.Union[str, dict]:

        '''
        self.ban_group_member(user_id='chat id or guid', chat_id='chat guid')
        
        PARAMETRS:
            1- self - self object
            2- user_id - target guid or chat id
            3- member_username - to do nt using user_id and to insert: '@username'
            4- link - to dont using chat_id and to insert: 'https://rubika.ir/+'
            5- action is actions type: 'Set', 'Unset'
        '''

        if (link):
                chat_id: str = Maker.check_link(link=link, client=self)
        if (member_username):
            user_id: str = Maker.check_link(link=member_username, client=self)

        return (
            GetData.api(version = self.api_version or '4',
                        method = 'banGroupMember', auth = self.auth,
                        data = {'group_guid': chat_id, 'member_guid': user_id, 'action': action}, proxy = self.proxy,
                        mode = self.city, platform = self.platform or 'android',))


    def add_group_members(
        self        :   'Client',
        chat_id     :   (str)   =   (None),
        user_ids    :   (list)  =  (None),
        *args,
        **kwargs,
        ) -> typing.Union[str, dict]:
            
            '''
            self.add_group_members(['user-guid', ], chat_id='group-guid')
            
            ## PARAMETERS:
                1- `self object`
                2- `chat_id is group guid`
                3- `user_ids is targets guid`
                
               
            '''

            return (
                GetData.api(
                    version='5',
                    method='addGroupMembers',
                    auth=self.auth,
                    data={
                        'group_guid'    :   chat_id,
                        'member_guids'  :   user_ids
                        },
                    mode=self.city,
                    proxy=self.proxy,
                    paltform='web',
                    )
                )


    def get_group_admin_members(
        self        :   ('Client'),
        chat_id     :   (str)   =   (None),
        username    :   (str)   =   (None),
        link        :   (str)   =   (None)
        ) -> typing.Union[str, dict]:

            '''
            self.get_group_admin_members('chat guid')
            
            PARAMETERS:
                -1 self is a self object
                -2 chat_id is chat guid
                -3 username is for dont using chat_id and to insert: '@username'
                -4 link is for dont using chat_id or username and to insert: 'https://rubika.ir/'
                - END :)
            '''
            
            if (username or link):
                chat_id: str = Maker.check_link(link=username or link, client=self)

            return (
                GetData.api(
                    version='5',
                    method='getGroupAdminMembers',
                    auth=self.auth,
                    data={
                        'group_guid'  :   (
                            chat_id
                            )
                        },
                    mode=self.city,
                    platform='web',
                    proxy=self.proxy
                    )
                )


    def set_group_default_access(
        self        :   ('Client'),
        access_list :   (list),
        chat_id     :   (str)   =   (None),
        username    :   (str)   =   (None),
        link        :   (str)   =   (None)
        ) -> typing.Union[str, dict]:
            
            
            '''
            from rubx import accesses
            self.set_group_default_access([accesses.SendMessages, ...], 'chat guid')
            
            PARAMETERS:
                1- self is a self object
                2- access_list is a list for users access
                3- chat_id is chat guid
                4- username is for dont using chat_id and to insert: '@username'
                5- link is for dont using chat_id or username and to insert: 'https://rubika.ir/username'
                - END :)
            '''

            if (username or link):
                chat_id: str = Maker.check_link(link=username or link, client=self)
            
            return (
                GetData.api(
                    version     =   self.api_version or '4',
                    method      =   'setGroupDefaultAccess',
                    auth        =   self.auth,
                    data        =   {
                        'access_list'   :   (access_list),
                        'group_guid'    :   (chat_id)
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'android'
                    )
                )


    def get_group_all_members(
        self        :   ('Client'),
        chat_id     :   (str)   =   (None),
        start_id    :   (str)   =   (None)
        ) -> typing.Union[str, dict]:
            
            '''
            self.get_group_all_members('chat guid')
            
            PARAMETERS:
                1- chat_id is chat guid (group type)
                2- start_id is for geting next member. and to get start_id: get key 'next_start_id'
                - END
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'getGroupAllMembers',
                    auth        =   self.auth,
                    data        =   {
                        'group_guid'    :   (chat_id),
                        'start_id'      :   (start_id)
                        },
                    proxy       =   self.proxy,
                    mode        =   self.city,
                    playform    =   self.platform or 'web'
                    )
                )


    def get_group_link(
        self    :   ('Client'),
        chat_id :   (str)
        ) -> typing.Union[str, dict]:
            
            '''
            self.get_group_link('group-guid')
            note: this method for admin group
            
            PARAMETERS:
                1- self is a self object
                2- chat_id is group guid
                - END
            '''
            
            return(
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getGroupLink',
                    data        =   {
                        'group_guid'    :   (chat_id)
                    },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def set_group_link(
        self    :   ('Client'),
        chat_id :   (str)   =   (None)
        ) -> typing.Union[str, dict]:
            
            '''
            `self.get_group_link('group-guid')`
            note: this method for admin group
            
            PARAMETERS:
                1- self is a self object
                2- chat_id is group guid
                - END
            '''
            
            return(
                GetData.api(
                    version     =   '5',
                    auth        =   self.auth,
                    method      =   'setGroupLink',
                    data        =   {
                        'group_guid'    :   (chat_id)
                    },
                    mode        =   self.city,
                    platform    =   'web',
                    proxy       =   self.proxy
                )
            )


    def set_group_timer(
        self    :   ('Client'),
        chat_id :   (str),
        time    :   (str),
        ) -> typing.Union[str, dict]:
            
            '''
            
            ## USE:
                `self.get_group_link('group-guid')`
                
            ## NOTE:
                this method for admin group
            
            ## PARAMETERS:
                - `self` is a self object
                - `chat_id` is group guid
                - `time` is a time for group timer
            ### END
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '4',
                    method      =   'editGroupInfo',
                    auth        =   self.auth,
                    data        =   {
                        'group_guid'            :   chat_id,
                        'slow_mode'             :   time,
                        'updated_parameters'    :   ['slow_mode']
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'android'
                )
            )


    def set_group_admin(
        self            :   ('Client'),
        access_list     :   (list)  =   (None),
        chat_id         :   (str)   =   (None),
        user_id         :   (str)   =   (None),
        action          :   (str)   =   ('SetAdmin')
        ) -> typing.Union[str, dict]:

            '''
            ### note: this method for admin a member in group.
            
            ## USE:
                `from rb.Accesses import Admin`
                `self.set_group_admin([Admin.SendMessages, ...], 'group guid', 'user guid')`
            
            ## PARAMETERS:
                - `self`: is a self object
                - `access_list`: is a list for member access
                - `chat_id`: is group guid
                - `user_id`: is member guid
                - `action`: is a action type. actions: 'SetAdmin', UnsetAdmin''
            '''

            return (
                GetData.api(
                    version     =   '5',
                    method      =   'setGroupAdmin',
                    auth        =   self.auth,
                    data        =   {
                        'group_guid'    :   chat_id,
                        'access_list'   :   access_list,
                        'action'        :   action,
                        'member_guid'   :   user_id
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   'web'
                    )
                )


    def join_group(
        self    :   ('Client'),
        link    :   (str)
        ) -> typing.Union[str, dict]:
            
            '''
            (this method for join to groups.)
            
            USE:
                self.join_group('https://rubika.ir/joing/...')
            PARAMETERS:
                1- self is a self object.
                2- link is a link rubika group.
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'joinGroup',
                    auth        =   self.auth,
                    data        =   {
                        'hash_link'   :   self.__link_scanner(link)
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def group_preview_by_join_link(
        self    :   ('Client'),
        link    :   (str)
        ) -> typing.Union[str, dict]:

            '''
            (this method for get group link info by join link)
            
            USE:
                self.group_preview_by_join_link('https://rubika.ir/joing/...')
            PARAMETERS:
                1- self is a self object.
                2- link is a link rubika group.
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   'groupPreviewByJoinLink',
                    auth        =   self.auth,
                    data        =   {
                        'hash_link'   :   self.__link_scanner(link)
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def leave_group(
        self    :   ('Client'),
        chat_id :   (str)
        ) -> typing.Union[str, dict]:
            
            '''
            this method for leave the group
            
            USE:
                self.leave_group('group guid') # g0...
            PARAMETERS:
                1- self is a self oobject
                2- chat_id is group guid
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or ('5'),
                    method      =   'leaveGroup',
                    auth        =   self.auth,
                    data        =   {
                        'group_guid'    :   (chat_id)
                        },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or ('web')
                )
            )


    def create_objcet_voice_chat(
        self    :   ('Client'),
        chat_id :   (str)
        ) -> typing.Union[str, dict]:

            '''
            self.start_voice_chat('chat-guid')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    method      =   f'create{Scanner.check_type(chat_id)}VoiceChat',
                    data        =   {
                        f'{Scanner.check_type(chat_id).lower()}_guid'   :   chat_id,
                        },
                    auth        =   self.auth,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web',
                    mode        =   self.city
                )
            )


    def set_object_voice_chat_setting(
        self            :   'Client',
        chat_id         :   (str),
        voice_chat_id   :   (str),
        title           :   (str)
        ) -> typing.Union[str, dict]:
            
            '''
            self.set_object_voice_chat_settin('chat-guid', 'id', 'the title')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   f'set{Scanner.check_type(chat_id)}VoiceChatSetting',
                    data        =   {
                        f'{Scanner.check_type(chat_id).lower()}_guid'   :   chat_id,
                        'voice_chat_id'                                 :   voice_chat_id,
                        'title'                                         :   title ,
                        'updated_parameters'                            :   ['title']
                    },
                    mode        =   self.city,
                    proxy       =   self.proxy,
                    platform    =   self.platform or 'web'
                )
            )


    def discard_object_voice_chat(
        self            :   'Client',
        chat_id         :   (str),
        voice_chat_id   :   (str),
        ) -> typing.Union[str, dict]:
            
            '''
            self.discard_object_voice_chat('chat-guid', 'id')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   f'discard{Scanner.check_type(chat_id)}VoiceChat',
                    data        =   {
                        f'{Scanner.check_type(chat_id).lower()}_guid'   :   chat_id,
                        'voice_chat_id'                                 :   voice_chat_id
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def set_action_chat(
        self    :   Client,
        chat_id :   str,
        action  :   str
        ) -> typing.Union[str, dict]:

            '''
            USE:
                self.set_action_chat('chat-guid', 'Pin')
            PARAMS:
                1- self is a self object
                2- chat_id is chat guid
                3- action is a action type. actions: 'Mute', 'Unmute' | 'Pin', 'Unpin'
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'setActionChat',
                    data        =   {
                        'object_guid'   :   chat_id,
                        'action'        :   action
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def join_group_voice_chat(
        self            :   Client,
        chat_id         :   str,
        voice_chat_id   :   str,
        self_object_guid:   str,
        sdp_offer_data  :   str) -> typing.Union[str, dict]:

            '''
            self.join_group_voice_chat('chat-guid', 'id', 'your-guid', ...)
            '''

            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'joinGroupVoiceChat',
                    data        =   {
                        'chat_guid'         :   chat_id,
                        'voice_chat_id'     :   voice_chat_id,
                        'sdp_offer_data'    :   sdp_offer_data,
                        'self_object_guid'  :   self_object_guid
                    },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def leave_group_voice_chat(
        self            :   Client,
        chat_id         :   str,
        voice_chat_id   :   str
        ) -> typing.Union[str, dict]:
            
            '''
            self.leave_group_voice_chat('chat-guid', 'id')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'leaveGroupVoiceChat',
                    data        =   {
                        'chat_guid'     :   chat_id,
                        'voice_chat_id' :   voice_chat_id
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def set_group_voice_chat_setting(
        self            :   Client,
        chat_id         :   str,
        voice_chat_id   :   str,
        **kwargs        :   (dict or str)
        ) -> typing.Union[str, dict]:
            
            '''
            self.set_group_voice_chat_setting('chat-guid', 'id', title='Hey')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'setGroupVoiceChatSetting',
                    data        =   {
                        'chat_guid'     :   chat_id,
                        'voice_chat_id' :   voice_chat_id,
                        'updated_parameters'    :   list(kwargs.keys())
                        }.update(kwargs),
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def edit_group_info(
        self                        :   Client,
        chat_id                     :   str,
        title                       :   str =   None,
        description                 :   str =   None,
        chat_history_for_new_members:   str =   None
        ) -> typing.Union[str, dict]:
            
            '''
            self.edit_group_info('chat-guid', 'name', 'bio')
            
            actions: Hidden, Visible
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'editGroupInfo',
                    data        =   {
                        'group_guid'        :   chat_id,
                        'title'             :   title,
                        'description'       :   description,
                        'updated_parameters':   ['title' if title else None, 'description' if title else None, 'chat_history_for_new_members' if chat_history_for_new_members else None]
                        }.update({'chat_history_for_new_members': chat_history_for_new_members} if chat_history_for_new_members else {}),
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def get_banned_chat_members(
        self    :   Client,
        chat_id :   str,
        start_id:   str =   None
        ) -> typing.Union[str, dict]:
            
            '''
            self.get_banned_group_members('chat-guid') # get 'next_start_id' from respone to star_id param
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '4',
                    auth        =   self.auth,
                    method      =   f'getBanned{Scanner.check_type(chat_id)}Members',
                    data        =   {
                        f'{Scanner.check_type(chat_id).lower()}_guid'        :   chat_id,
                        }.update({'start_id' : start_id} if start_id else {}),
                    mode        =   self.city,
                    platform    =   self.platform or 'android',
                    proxy       =   self.proxy
                )
            )


    def get_group_mention_list(
        self            :   Client,
        chat_id         :   str =   None,
        search_mention  :   str =   None
        ) -> typing.Union[str, dict]:
            
            '''
            self.get_group_mention_list('chat-guid')
            '''
            
            return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'getGroupMentionList',
                    data        =   {
                        'group_guid'    :   chat_id,
                        'search_mention':   search_mention
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def delete_no_access_group_chat(self: Clean, chat_id: str) -> typing.Union[str, dict]:
        
        '''
        self.delete_no_access_group_chat('group-guid')
        '''

        return (
                GetData.api(
                    version     =   self.api_version or '5',
                    auth        =   self.auth,
                    method      =   'deleteNoAccessGroupChat',
                    data        =   {
                        'group_guid'    :   chat_id
                        },
                    mode        =   self.city,
                    platform    =   self.platform or 'web',
                    proxy       =   self.proxy
                )
            )


    def get_common_groups(self, user_id: str) -> typing.Union[str, dict]:
        '''
        `self.get_common_groups('u0...')`
        '''
        
        return (GetData.api(version = self.api_version or '4', auth = self.auth,
                            method = 'getCommonGroups', data = {'user_guid': user_id},
                            mode = self.city, platform = self.platform or 'android',
                            proxy = self.proxy))


    def get_group_online_count(self, chat_id: str) -> typing.Union[str, dict]:
        '''
        `self.get_group_online_count('g0...')`
        '''

        return (GetData.api(version = self.api_version or '4', auth = self.auth,
                            method = 'getGroupOnlineCount', data = {'group_guid': chat_id},
                            mode = self.city, platform = self.platform or 'android',
                            proxy = self.proxy))

__all__ = ['GroupMethods']


#!/bin/python

from .connection import GetData
from .sessions import RubikaClient as Client


class Method(object):

    @classmethod
    def from_json(cls, session: str, method_name: str,
                  *args, **kwargs) -> (dict):

        '''
        # this is a method to use custom method on rubika client
        
        Method('session', 'SendMessage', chat_id='u0...', text='Hey!')
        
        '''

        data: dict = {}
        
        assert list(map(lambda key: data.update({key: kwargs.get(key)}, list(kwargs.keys()))))

        return (
            GetData.api(
                version     =   '5',
                method      =   method_name[0].lower() + method_name[1:],
                auth        =   session,
                data        =   data,
                proxy       =   {'http': 'http://127.0.0.1:9050'},
                platform    =   'rubx',
                mode        =   'mashhad'
            )
        )


#!/bin/python

class SessionError(OSError):
    pass

class MainError(Exception):
    pass

class NotREGISTERED(IOError):
    pass

class InvalidInput(IOError):
    pass

class TooREQUESTS(IOError):
    pass

class InvalidAUTH(IOError):
    pass

class ConnectError(OSError):
    pass

class ClientError(Exception):
    pass

class ServerError(Exception):
    pass

class ApiHasNotResponse(Exception):
    pass

class APIError(Exception):
    pass

class BotApiTimeOutError(Exception):
    pass

__all__ = ['SessionError', 'MainError', 'NotREGISTERED',
           'InvalidInput', 'TooREQUESTS', 'InvalidAUTH',
           'ConnectError', 'ClientError', 'ServerError',
           'APIError', 'BotApiTimeOutError', 'ApiHasNotResponse']


#!/bin/python
# rubika self client models

from pydantic import BaseModel
import typing


class File(BaseModel):
    file_id: str = None
    file_name: str = None
    dc_id: str = None
    access_hash_rec: str = None
    type: str = None
    time: str = None
    text: str = None
    is_edited: bool = None
    is_deleted: bool = None
    size: str = None


class Message(BaseModel):
    message_id: str = None
    text: str = None
    reply_to_message_id: str = None
    time: str = None
    is_edited: bool = None
    type: str = None
    author_type: str = None
    author_object_guid: str = None
    object_guid: str = None
    last_message_id: str = None
    group_guid: str = None
    channel_guid: str = None
    user_guid: str = None
    first_name: str = None
    last_name: str = None
    username: str = None
    member_guid: str = None
    file_inline: typing.Optional[File] = None
    count_seen: str = None


class Bot(BaseModel):
    original: str = None
    message: dict = None
    messages: list = None
    chats: list = None
    in_chat_members: list = None
    message_id: str = None
    text: str = None
    reply_to_message_id: str = None
    time: str = None
    is_edited: bool = None
    type: str = None
    author_type: str = None
    author_object_guid: str = None
    object_guid: str = None
    last_message: typing.Optional[Message] = None
    abs_object: typing.Optional[Message] = None
    count_seen: str = None
    group_guid: str = None
    channel_guid: str = None
    user_guid: str = None
    first_name: str = None
    last_name: str = None
    username: str = None
    title: str  = None
    description: str = None
    file_inline: typing.Optional[File] = None
    last_message_id: str = None
    next_start_id: str = None


class Result(BaseModel):
    status: str = None
    status_det: str = None
    data: typing.Optional[Bot] = None


class Attrs(object):

    @classmethod
    def create(cls, __names: dict, __bases: tuple = (Bot, ), /,
               action: str = 'bot', *args, **kwargs) -> typing.Union[Bot, Message, File, Result, object]:
        
        
        '''
        data = {'chat_id': ..., 'message_id': ...}
        
        print(Attr.create(data))
        '''

        __names.update({'original': 'updates'})
        
        for name in __names.keys():
            setattr(Bot, name, __names.get(name))
            #type(name, (Bot, ), {'__name__': name, name: __names.get(name)})

        for attr, name in zip(__names.keys(), map(lambda name: getattr(Bot, name), __names.keys())):
            if not name:
                delattr(Bot, attr)

        return Message(**__names) if action.__eq__('message') else Bot(**__names) if action.__eq__('bot') else File(**__names) if action.__eq__('file') else Result(**__names)


__all__ = ['Attrs']


#!/bin/python
# main object - `rb` is main package to importing

import os, sys

if __name__.__eq__('__main__').__and__(__package__.__eq__(None) or __package__.__eq__('')):
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

try:
    import tinytag, Crypto, pydantic
except ModuleNotFoundError:
    from .extensions import PyPi
    PyPi().installation(['tinytag', 'pycryptodome', 'pydantic'])

import logging, platform, typing, difflib, inspect

from datetime           import  datetime
from json               import  dumps, loads
from random             import  choice, randint, sample
from re                 import  findall, search, sub, compile
from time               import  sleep, gmtime, localtime
from warnings           import  warn
from requests           import  get, post, session
from abc                import  ABC
from .exceptions        import  *
from .extensions        import  *
from .events            import  *
from .crypto            import  Encryption
from .connection        import  GetData, Urls, Connection, Make
from .storage           import  SQLiteSession
from .clients           import  *
from .UserMethods       import  UserMethods
from .GroupMethods      import  GroupMethods
from .ChannelMethods    import  ChannelMethods
from .rubino            import  RubinoClient
from .BotMethods        import  BotMethods
from .methods           import  Method
from .models            import  Attrs
from .parser            import  (
    MessageEmpty, MessageEntityBold, MessageEntityCode,
    MessageEntityItalic, MessageEntityHashtag, MessageEntityMention,
    MessageEntityMentionName, MessageEntityPre, MessageEntityStrike,
    MessageEntityTextUrl, MessageEntityUnderline, MessageEntityUnknown,
    MessageEntityUrl, Metas, Tags, MetaDataLoader, MarkDown
    )

if typing.TYPE_CHECKING:
    from . import RubikaClient, UserMethods

try:
    from . import __file__ as base_file
    from . import __name__ as base_log
    from . import __package__ as base_pkg

except ImportError:
    base_file, base_log, base_pkg = locals().get('__file__') or __doc__, locals().get('__name__'), locals().get('__package__')


class Version(str):
    __version__ =   '10.5.5'
    __author__  =   'saleh'
    __lisense__ =   'MIT'
    __module__  =   'rubx'


class RubikaClient(ABC):

    __slots__ = ('self', 'session', 'chat_id',
                 'username', 'app', 'phone_number',
                 'device', 'proxy', 'your_name',
                 'city', 'banner', 'creator_channel_open',
                 'platform', 'api_version', 'headers',
                 'timeout', 'check_update', 'lang_code',
                 'base_logger', 'check_session', 'api_client',
                 'return_data_an_dict')

    __base_class__ = 'RubikaClient'

    def __str__(self, *args) -> (str):
        return dumps({'session': SQLiteSession(self.auth).information()}, indent=2) if self.auth else dumps({'__base_class__': __base_class__}, indent=2)

    def __init__(
        self                :   ('Client'),
        session             :   (str)                               =   (None),
        chat_id             :   (str)                               =   (None),
        username            :   (str)                               =   (None),
        app                 :   (str)                               =   ('rubx'),
        phone_number        :   (str)                               =   (None),
        device              :   (dict)                              =   (Device.DEFAULT_DEVICE),
        proxy               :   (dict)                              =   {'http': 'http://127.0.0.1:9050'},
        your_name           :   (str)                               =   (False),
        city                :   (str)                               =   ('mashhad'),
        banner              :   (bool)                              =   (False),
        creator_channel_open:   (bool)                              =   (False),
        platform            :   (str)                               =   (None),
        api_version         :   (typing.Union[str, int])            =   (None),
        headers             :   (typing.Union[dict, str, list])     =   (None),
        timeout             :   ((typing.Union[int, str]))          =   (5),
        check_update        :   (typing.Optional[bool])             =   (False),
        lang_code           :   (str)                               =   ('fa'),
        base_logger         :   (typing.Union[str, logging.Logger]) =   (None),
        check_session       :   (bool)                              =   (False),
        api_client          :   (str)                               =   (None),
        return_data_action  :   (str)                              =   ('dict'),
        *args,
        **kwargs
        ) -> (None):

        '''
        ## IMPORT:
            - from rb import RubikaClient

        ## USE:
        
            `client = RubikaClient('session-key', 'u0...', 'username', 'rubx', your_name='saleh', banner=True, creator_channel_open=True, platform='rubx', api_version='5', timeout=5, proxy={'socks5':'http://127.0.0.1:9050'}, headers={'user-agent':...}, api_client='https://messengerg56c.iranlms.ir:80', ...)`

        ## EXAMPLES:
            
            `with RubikaClient(...) as client:`
                `client.send_message(...)`
            
            # ... or ...
            
            `client = RubikaClient('abcdefghijklmnopqrstuvwxyzzabcde')`
            `client.session = '' # len: 32 chars`
            
            `import socks`
            `client.proxy   = {socks.SOCKS5:'127.0.0.1:9050'}`
            
            `def run(callable, params) -> dict:`
                `return callable(**params)`

            ```
            print(
                run(
                    client.send_message,
                    dict(
                        chat_id='chat-guid',
                        text='Hey! this message from rubx lib.'
                        )
                    )
                )
            ```

        ## PARAMETERS:
        
            - 1- `self`: is a self obejct
            - 2- `session`: is account key [auth]
            - 3- `chat_id`: is your guid account
            - 4 - `username`: is your username account
            - 5 - `app`: is from app name
            - 6 - `phone_number`: is for using lib with phone_number and geting account key
            - 7 - `device`: is your account device for use token or thumbinline or ...
            - 8 - `proxy`: set proxy http to requests
            - 9 - `your_name`: is for save info in a file.
            - 10 - `city`: is for your countery and city for using client server.
            - 11 - `banner`: is a boolean for print banner
            - 12 - `creator_channel_open`: is for joining your account in creator channel
            - 13 - `platform`: is for using user platform. examples: `rubx` or `web` or `android`
            - 14 - `api_version`: is for using api mode: `5` (for web and rubx) `4` (for rubika app [andorid]) `3` (for m.rubka.ir)
            - 15 - `headers`: is for set header to requests
            - 16 - `timeout`: is for requests timeout
            - 17 - `check_update`: is for checking lib new version
            - 18 - `lang_code`: to app lang code. `en`, `fa`, ...
            - 19 - `base_logger`: is for `__name__`
            - 20 - `check_session`: return a dict type to checking session [AUTH]
            - 21 - `api_client`: to set a server for requests
            - 22 - `return_data_an_dict`: return the json data to str(ordered with dumps) or dict(normal)
        
        
        ## `what is [guid]?` : guide unique identifier | chat id : group guid, channel guid and others ...
        ## `what is [session] or [auth]?`: is your session id (api key) for account.
        ## what is [object]?: the 'object' is 'chat'


        ## ERRORS | EXCEPTIONS:
        
            - `NotREGISTERED`: your session false
            - `InvalidInput` : the request is blocked due to lack of access or wrong parameters.
            - `TooREQUESTS`  : this request is not possible due to the high number of requests. U should try again after a few moments!
            - `InvalidAUTH`‍  : the key or chat id or input entered is invalid, please check the information and then try again.
            - `ConnectError` : the server is down due to many requests and does not respond. You can use 'timeout' .
            - `ClientError`  : the client server is incorrect.
            - `ServerError`  : the server is unable to respond for any reason.
            - `SessionError` : your session is invalid.
            
            #### bypass them using (`try` & `except`) exception and solve them.

        to show public methods:
        ----------------------
            dir(RubikaClient)

        to show method comment [doc]:
        ----------------------------
            help(RubikaClient) or a attrbute

        for getting file thumbnail:
        --------------------------
            `get file bytes and encode to base64` `:` `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`

        ### star the repo: https://github.com/mester-root/rubx
        
        '''

        (self.app, self.proxy, self.enc,
            self.city, self.platform, self.api_version,
            self.headers, self.username, self.chat_id,
            self.handling, self.phone, _log,
            self.timeout, self.lang_code, self.device,
            self.check_session, self.api_client) = (app, proxy, Encryption(session),
                                                    city, platform, api_version,
                                                    headers, username, chat_id,
                                                    {}, phone_number, logging.getLogger(__name__),
                                                    timeout, lang_code, device,
                                                    check_session, api_client)
        
        Infos.citys.append(city)
        Infos.proxys.append(proxy)
        GetData.url: str = (api_client) # TODO: set other params 
        # TODO: set all params in __init__ object
        
        if banner:
            assert list(map(lambda character: (print(character, flush=True, end=''), sleep(0.01)), f'\n\033[0m< \033[31mrubx \033[0m> \033[36m | \033[31mstarted in \033[0m{str(datetime.datetime.now())}\033[31m| \033[0m{Version.__version__}\n'))
        
        if your_name:
            open('session_info.sty', 'w+').write('name: '+your_name+'\ntime started: '+str(datetime.datetime.now())+f'\key: {session}'+'\nyour ip: '+str(get('https://api.ipify.org').text))
        
        if session:
            if session.__len__() != 32:
                raise SessionError('Your `session` is invalid - the number of characters in your session is different from 32')
            
            self.auth: (str) = (session)
            Infos.auth_.append(session)
        
        elif phone_number:
            Login.SignIn(phone_number, self)
            
        else:
            warn('SessionWarning: please insert session key or phone_number in object')

        if creator_channel_open:
            # self.join_channel_action('')
            # __import__('webbrowser').open('https://rubika.ir/TheClient')
            pass

        if (app.__eq__('rubx')):

            try:
                open(f'{session}.sty', 'r')
            except FileNotFoundError:
                open(f'{session}.sty', 'w').write(session)
                database = SQLiteSession(session)
                database.insert(self.phone, session, self.chat_id, self.api_client or Urls.get_url())

        if check_update:
            UpToDate(Version.__version__, 'https://raw.githubusercontent.com/Mester-Root/rubx/main/rb/version.sty').user
        
        if check_session:
            self.check_session: dict = _Top(session).detecting()
        
        if isinstance(base_logger, str):
            base_logger: str = logging.getLogger(base_logger)
        
        elif not isinstance(base_logger, logging.Logger):
            
            try:
                base_logger: str = base_log
            except Exception: ...

        class Loggers(dict):
            def __missing__(self, key) -> ...:
                if key.startswith('rb.'):
                    key = key.split('.', maxsplit=1)[1]

                return base_logger.getChild(key)

        self._log = Loggers()
        # self.__name__ = base_logger
        
        Connection.timeout: int = timeout or 5
        clients.web.update({'lang_code': self.lang_code or 'en'}) # TODO: to set lang_code in rubx and android clients
        
        if isinstance(return_data_action, (str, tuple, list)):

            if isinstance(return_data_action, (list, tuple)):
                Make.action, Make.type = return_data_action

            elif isinstance(return_data_action, str) and any(return_data_action.__contains__(action) for action in ('dict', 'str', 'object')):
                Make.action: str = return_data_action

        if headers:
            Connection.headers = headers
        if api_version:
            Connection.api_version = api_version
        if self.platform:
            Connection.platform = platform
        if city:
            Connection.city = city
        if proxy:
            Connection.proxy = proxy

    def __dir__(self):
        return RubikaClient.__dir__(self)

    def __reduce__(self):
        return object.__reduce_ex__

    def __missing__(self, path):
        return path

    def __reduce_ex__(self):
        return object.__reduce_ex__

    def __init_subclass__(self):
        return object.__init_subclass__

    def __subclasshook__(self):
        return object.__subclasscheck__

    def __sizeof__(self):
        return object.__sizeof__

    def __hash__(self):
        return hash(RubikaClient)

    def __call__(self) -> (None):
        pass

    def __enter__(self):
        return (self)

    def __exit__(self, *args,
                 **kwargs) -> (None):
        pass

    @property
    def start(self): # TODO: set runner with start func
        pass

    @property
    def get_session(self) -> (str):
        '''
        getting session key - auth
        '''

        return self.auth

    @property
    def get_storage(self) -> (str):

        '''
        get local storage - SqliteSession - info's
        '''

        return dumps({'session': SQLiteSession(self.auth).information()}, indent=2)
    
    def get_me(self) -> str:
        '''
        get my sessions
        '''

        return self.get_storage

    @classmethod
    def run(cls, func: object,
            *args, **kwargs) -> None:
        
        '''
        running sync methods
        '''
        
        __import__('_thread').start_new_thread(func, *args, **kwargs)


Client = (RubikaClient)

class SetClient(
    RubikaClient,
    UserMethods,
    GroupMethods,
    ChannelMethods
    ):
    pass

class WebSocket(object):

    def __init__(self, session) -> (None):
        self.auth: str = session
        self.enc: Encryption = Encryption(session)

    def __enter__(self) -> (object):
        return self

    def __exit__(self, *args, 
                 **kwargs) -> (None):
        pass

    def __on_open(self, ws, api_version='4') -> (None):
        ws.send(dumps({'api_version': api_version, 'auth': self.auth, 'method': 'handShake'}))
    
    def __on_send(self, ws) -> (None):
        ws.send(dumps({}))

    def __on_message(self, message) -> (dict):
        
        try:

            parsedMessage: dict = loads(message)
            return {'type': parsedMessage.get('type') or 'messenger', 'data': loads(self.enc.decrypt(parsedMessage.get('data_enc')))}

        except Exception:
            return {'type': '', 'data': ''}

    def __connector(self, ws: object, mode: str = '1') -> (None):
        try:
            if mode.__eq__('1'):
                self.__on_open(ws, '4')
            else:
                self.__on_send(ws)
        except Exception:
            pass
    
    @property
    def connection(self) -> (None):

        from websocket import create_connection

        ws, count = create_connection('wss://jsocket4.iranlms.ir:80'), 0
        self.__connector(ws)

        while 1:

            try:
                for res in [ws.recv()]:
                    
                    count: int = count.__add__(1)

                    if ((count / 2).__round__()).__eq__(40): # 80
                        self.__connector(ws, '2')

                    elif ((count / 2).__round__()).__eq__(60): # 120
                        self.__connector(ws, '1')
                        count: int = 0

                    if str(res).__len__().__ne__(33):
                        yield self.__on_message(res)

            except Exception:
                pass


class EventBuilder(Client):
    
    def __str__(self) -> (str):
        return self.jsonify(indent=2)

    def __getattr__(self, name) -> (list):
        return self.find_keys(keys=name)

    def __setitem__(self, key, value) -> (None):
        self.original_update[key] = value

    def __getitem__(self, key):
        return self.original_update[key]

    def __lts__(self, update: list) -> (list):
        for index, element in enumerate(update):
            if isinstance(element, list):
                update[index] = self.__lts__(update=element)

            elif isinstance(element, dict):
                update[index] = EventBuilder(update=element)

            else:
                update[index] = element
        return update

    def __init__(self, update: dict = None) -> (None):

        '''
        # get all keys from a dictionary
        '''

        self.original_update = update

    def to_dict(self) -> (dict):
        return self.original_update

    def jsonify(self, indent=None) -> (str):
        
        result = self.original_update
        result['original_update'] = 'dict{...}'
        
        return dumps(
            result,
            indent=indent,
            ensure_ascii=False,
            default=lambda value: str(value)
            )

    def find_keys(self, keys: list,
                  original_update: str = None) -> (list):

        if not original_update:
            original_update = self.original_update

        if not isinstance(keys, list):
            keys = [keys]

        if isinstance(original_update, dict):
            for key in keys:
                try:
                    update = original_update[key]
                    if isinstance(update, dict):
                        update = EventBuilder(update=update)

                    elif isinstance(update, list):
                        update = self.__lts__(update=update)

                    return update

                except KeyError:
                    pass
            original_update = original_update.values()

        for value in original_update:
            if isinstance(value, (dict, list)):
                try:
                    return self.find_keys(keys=keys, original_update=value)

                except AttributeError:
                    pass

        raise AttributeError(f'Struct object has no attribute {keys}')

    @property
    def action(self):
        return self.find_keys(keys=['author_type'])


    @property
    def type(self):
        try:
            return self.find_keys(keys=['type', 'author_type'])

        except AttributeError:
            pass

    @property
    def raw_text(self):
        try:
            return self.find_keys(keys='text')

        except AttributeError:
            pass

    @property
    def message_id(self):
        try:
            return self.find_keys(keys=['message_id',
                                        'pinned_message_id'])
        except AttributeError:
            pass

    @property
    def reply_message_id(self):
        try:
            return self.find_keys(keys='reply_to_message_id')

        except AttributeError:
            pass

    @property
    def find_rubika_channel_post(self):
        return findall(r'(rubika\.ir\/\w{4,25}\/\w{15})', self.raw_text)

    @property
    def find_rubika_private_link(self):
        return findall(r'(rubika\.ir\/join[c,g]\/\w{32})', self.raw_text)

    @property
    def find_chanel_private_link(self):
        return findall(r'(rubika\.ir\/joinc\/\w{32})', self.raw_text)

    @property
    def find_group_link(self):
        return findall(r'(rubika\.ir\/joing\/\w{32})', self.raw_text)

    @property
    def find_rubika_link(self):
        return findall(r'(rubika\.ir\/\w{4,25})', self.raw_text)

    @property
    def find_atsign(self):
        return findall(r'\@\w{4,25}', self.raw_text)

    @property
    def find_url(self):
        return findall(r'(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,}|www\.[a-zA-Z0-9]+\.[^\s]{2,}|[a-zA-Z0-9]+\.[^\s]{2,})', self.raw_text)

    @property
    def is_rubika_channel_post(self):
        return (search(r'(rubika\.ir\/\w{4,25}\/\w{15})', self.raw_text)).__bool__()

    @property
    def is_rubika_private_link(self):
        return (search(r'(rubika\.ir\/join[c,g]\/\w{32})', self.raw_text)).__bool__()

    @property
    def is_chanel_private_link(self):
        return (search(r'(rubika\.ir\/joinc\/\w{32})', self.raw_text)).__bool__()

    @property
    def is_group_link(self):
        return (search(r'(rubika\.ir\/joing\/\w{32})', self.raw_text)).__bool__()

    @property
    def is_rubika_link(self):
        return (search(r'(rubika\.ir\/\w{4,25})', self.raw_text)).__bool__()

    @property
    def is_atsign(self):
        return (search(r'\@\w{4,25}', self.raw_text)).__bool__()

    @property
    def is_url(self):
        return (search(r'(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s]{2,}|www\.[a-zA-Z0-9]+\.[^\s]{2,}|[a-zA-Z0-9]+\.[^\s]{2,})', self.raw_text)).__bool__()
    
    @property
    def is_admin(self):
        return any(user.get('member_guid').__eq__(self.author) for user in SetClient(self.auth).get_group_admin_members(self.object_guid).get('data').get('in_chat_members'))
    
    @property
    def is_user(self):
        return self.action.__eq__('User')
    
    @property
    def is_group(self):
        return self.type.__eq__('Group')

    @property
    def is_channel(self):
        return self.type.__eq__('Channel')

    @property
    def is_private(self):
        return self.type.__eq__('User')
    
    @property
    def is_personal(self):
        return self.is_private
    
    @property
    def is_bot(self):
        return self.type.__eq__('Bot')
    
    @property
    def is_service(self):
        return self.type.__eq__('Service')
    
    @property
    def object_guid(self):
        try:
            return self.find_keys(keys=['group_guid', 'object_guid',
                                        'channel_guid', 'user_guid',
                                        'bot_guid', 'service_guid'])
        except AttributeError:
            pass

    @property
    def author(self):
        try:
            return self.find_keys(keys=['author_object_guid'])

        except AttributeError:
            pass
        
    @property
    def author_object_guid(self):
        try:
            return self.author_object_guid

        except AttributeError:
            pass
    
    def guid_type(self, chat_id: str) -> str:
        if isinstance(chat_id, str):
            return Scanner.check_type(chat_id)
    
    def finder(self, filters) -> (object):
        
        if 'group_guid' in filters:
            return self.group_guid if not 'message' in self.original_update.keys() else self.message.group_guid
        elif 'channel_guid' in filters:
            return self.channel_guid if not 'message' in self.original_update.keys() else self.message.channel_guid
        elif 'user_guid' in filters:
            return self.user_guid if not 'message' in self.original_update.keys() else self.message.user_guid
        elif 'object_guid' in filters:
            return self.object_guid if not 'message' in self.original_update.keys() else self.message.object_guid
        else:
            return self.author_object_guid if not 'message' in self.original_update.keys() else self.message.author_object_guid
    
    def pin(self, chat_id: str = None,
            message_id: str = None) -> (dict):

        return SetClient(self.auth).set_pin_message(chat_id = chat_id or self.find_keys(keys=['object_guid', 'author_object_guid', 'channel_guid', 'group_guid', 'user_guid', 'bot_guid', 'service_guid']), message_id = message or self.message_id, action='Pin')
    
    def unpin(self, chat_id: str = None,
              message_id: str = None) -> (dict):

        return SetClient(self.auth).set_pin_message(chat_id = chat_id or self.find_keys(keys=['object_guid', 'author_object_guid', 'channel_guid', 'group_guid', 'user_guid', 'bot_guid', 'service_guid']), message_id = message or self.meesage_id, action='Unpin')

    def seen(self, chat_id: str = None,
             message: str = None) -> (dict):
        
        return SetClient(self.auth).seen_chats({chat_id or self.find_keys(keys=['object_guid', 'author_object_guid', 'channel_guid', 'group_guid', 'user_guid', 'bot_guid', 'service_guid']): message or self.message_id if not 'message' in self.original_update.keys() else self.message.message_id})
    
    def reply(self, *args, **kwargs) -> (dict):
        
        return SetClient(self.auth).send_message(*args, **kwargs)
    
    def respond(self, text: str, 
                action: str = 'author_object_guid') -> (dict):

        return SetClient(self.auth).send_message(text, self.finder(action), reply_to_message_id=self.message_id if not 'message' in self.original_update.keys() else self.message.message_id)
    
    def send(self, text,
             action: str = 'author_object_guid'):
        return SetClient.send_message(text, self.finder(action))
    
    def edit(self, text: str, chat_id: str = None,
             message_id: str = None, action: str = 'author_object_guid', *args, **kwargs) -> (dict):

        return SetClient(self.auth).edit_message(message_id = message_id or self.message_id, text=text, chat_id = chat_id or self.finder(action), *args, **kwargs)
    
    def forwards(self, to: str,
                 _from: str = None, messages: list = None) -> (dict):
        
        return SetClient(self.auth).forward_messages(_from or self.find_keys(keys=['object_guid', 'author_object_guid', 'channel_guid', 'group_guid', 'user_guid', 'bot_guid', 'service_guid']), messages or [self.message_id], to)
    
    def download(self, chat_id: str = None,
                 message: str = None, name: str = None, *args, **kwargs) -> (dict):

        return SetClient(self.auth).get_file('message' if message else '', True, save_as=name, chat_id = chat_id or self.find_keys(keys=['object_guid', 'author_object_guid', 'channel_guid', 'group_guid', 'user_guid', 'bot_guid', 'service_guid']), message = message or self.find_keys(keys=['messaage_id']))

    def delete(self, chat_id: str = None,
               messages: list = None, action: str = 'object_guid', *args, **kwargs) -> (dict):
        
        return SetClient(self.auth).delete_messages(messages or [self.message_id], chat_id or self.finder(action))

    def save(self, action: str = 'author_object_guid') -> (dict):

            return SetClient(self.auth).forward_messages(self.finder(action), [self.message_id], self.chat_id or self.user_guid)

    def activity(self, chat_id: str = None,
                 action: str = 'author_object_guid') -> (dict):
        return SetClient(self.auth).send_chat_activity(self.finder(action), 'is typing...')


class NewMessage(Client):
    
    def __init__(
        self,
        func        :   object,
        filters     :   list  =   None,
        pattern     :   str   =   None,
        commands    :   dict  =   None, 
        gets        :   tuple =   None,
        handle_name :   str   =   'handshake',
        *args
        ) -> (None):
        
        '''
        ## PARAMS:
            - `filters  = ['u0...', ]`
            - `pattern  = '(?) hi \S+'`
            - `commands = {'/start': 'Hey!'}`
            - `gets     = ('/start', '/info')`
        '''

        super().__init__(func, filters, pattern,
                         commands, gets, handle_name)
        (self.pattern, self.filters, self.gets,
         self.handle_name, self.commands, self.func) = (pattern, filters, gets,
                                                        handle_name, commands, func)


    def builder(self) -> iter:
        
        if not isinstance(self.filters, list):
            self.filters: list = [self.filters]
        
        for message in self.func(self.handle_name, get_messages=True, chat_ids=self.filters):
            
            if self.pattern:
                if not isinstance(self.pattern, str):
                    raise ValueError('oh pattern param is not string type.')
                if search(self.pattern, message.get('text') or message):
                    update = EventBuilder(message)
                    yield update

            elif self.commands:
                
                if not isinstance(self.commands, dict):
                    raise ValueError('oh commands param is not dictionary type.')
                    
                if any(message.get('text').__contains__(cmd) for cmd in list(self.commands.keys())):
                    update = EventBuilder(message)
                    yield update

            elif self.gets:
                
                if not isinstance(self.gets, tuple) or isinstance(self.gets, list):
                    raise ValueError('oh gets param is not a tuple or list type.')
                
                if any(cmd in message for cmd in self.gets):
                    update = EventBuilder(message)
                    yield update

            else:
                update = EventBuilder(message)
                yield update


class Handler(Client):

    def __init__(self, starting: bool = True,
                 *args, **kwargs) -> (None):

        '''
        # Handler | Events


        ## Example:
            from rb import Handler, NewMessage

            with Handler(...) as client:
                client.on(NewMessage(client.handle, handle_name='ChatsUpdates').builder())
                def update(event):
                    pass

        ## Responses:
            - `HandShake`:

                message_updates:

                    {
                        'message_id': ...,
                        'message': {
                            'text': ...,
                            'message_id': ...,
                            'action': ...,
                            'type': ...,
                            'author_object_guid': ..., 
                            },
                        'user_guid': ...,
                        }

                chat_updates:
                    ...

                show_notifications:
                    ...

            - `ChatsUpdates`:
                {
                    'message_id': ...,
                    'text': ...,
                    'author_object_guid' ...,
                    'author_type': ... ,
                    'type': ...
                    }

            - `MessagesUpdates`:
                {
                    'message_id': ...,
                    'text': ...,
                    'author_object_guid' ...,
                    'author_type': ... ,
                    'type': ...
                    }
        '''
        
        super().__init__(starting, *args, **kwargs)
        self.starting = starting

    def __enter__(self):
        return self

    def __exit__(self, *args,
                 **kwargs) -> (None):
        pass

    def __appender(self, msg_id: str = None,
                    action: str = 'edit'):
        
        if action == 'edit':
            if not msg_id in open(self.auth+'-ids.sty', 'r').read():
                open(self.auth+'-ids.sty', 'a+').write(msg_id+'\n')
        else:
            open(self.auth+'-ids.sty', 'w')
    
    def handle(
            self,
            method              :   str                         =   'ChatsUpdates',
            get_chats           :   bool                        =   True,
            get_messages        :   bool                        =   True,
            chat_ids            :   str                         =   None,
            author_guid         :   str                         =   None,
            pattern             :   typing.Union[tuple, list]   =   None,
            show_notifications  :   bool                        =   False
            ) -> (dict):

        '''
        `method`: methods: `ChatsUpdates`, `MessagesUpdates`, `HandShake`
        `get_chats`: the chat updates
        `get_messages`: the message updates
        `chat_ids`: chat filter
        `author_guid`: author_filter
        `pattern`: the pattern is for get message filter: `('^\w{1}start', 'Hey! from rubx lib')`
        '''

        if not method or not isinstance(method, str) or not any(word in method.lower() for word in ('chatsupdates', 'messagesupdates', 'handshake', 'socket')):
            method: str = 'HandShake'

        if (method.lower().__contains__('handshake') or method.lower().__contains__('socket')):

            for msg in self.hand_shake():

                if msg.get('type') == 'messenger':

                    res: dict = msg.get('data')

                    if get_chats and get_messages:
                        if not show_notifications:
                            res.pop('show_notifications')
                        yield res

                    elif get_messages:

                        for i in res.get('message_updates'):

                            if pattern:

                                if not isinstance(pattern, list) or isinstance(pattern, tuple):
                                    raise ValueError('pattern not a tuple or list type.')
        
                                if search(pattern[0], i.get('message').get('text') or ''):
                                    i.update({'pattern': pattern[1]})
                                    yield i
                            else:
                                yield i

                    elif get_chats:
                        for i in res.get('chat_updates'):
                            yield i

                    elif show_notifications:
                        for i in res.get('show_notifications'):
                            yield i

                    else:
                        res.update(res.get('message_updates'))
                        res.update(res.get('chat_updates'))
                        res.update(res.get('show_notifications'))
                        yield res

        elif (method.lower().__contains__('chatsupdates')):

            while (1):
                
                try:
                    for msg in SetClient(self.auth).get_chats_updates().get('data').get('chats'):
                        
                        if not msg.get('last_message').get('message_id') in open(self.auth+'-ids.sty', 'r').read():
                            
                            if chat_ids:
                                
                                if not isinstance(chat_ids, list):
                                    chat_ids: list = [chat_ids]
                                    
                                if msg.get('last_message').get('object_guid') in chat_ids or msg.get('last_message').get('author_object_guid') in chat_ids:
                                    
                                    if pattern:
                                        if search(pattern[0], msg.get('last_message').get('text')):
                                            
                                            msg.update(msg.get('last_message'))
                                            msg.update(msg.get('abs_object') or {})
                                            del msg['last_message']
                                            del msg['abs_object']
                                            msg.update({'pattern': pattern[1]})
                                            
                                            yield msg
                                    else:
                                        
                                        msg.update(msg.get('last_message'))
                                        msg.update(msg.get('abs_object') or {})
                                        del msg['last_message']
                                        del msg['abs_object']
                                        
                                        yield msg
                            else:
                                
                                if pattern and isinstance(pattern, tuple) or isinstance(pattern, list):
                                    if search(pattern[0], msg.get('last_message').get('text')):
                                        
                                        msg.update(msg.get('last_message'))
                                        msg.update(msg.get('abs_object') or {})
                                        msg.update({'pattern': pattern[1]})
                                        del msg['last_message']
                                        del msg['abs_object']
                                        
                                        yield msg
                                else:
                                    
                                    msg.update(msg.get('last_message'))
                                    msg.update(msg.get('abs_object') or {})
                                    del msg['last_message']
                                    del msg['abs_object']
                                    
                                    yield msg
                            
                        self.__appender(msg.get('last_message').get('message_id'), 'edit')
                except Exception:
                    ...

        elif (method.lower().__contains__('messagesupdates')):

            while (1):
                
                try:
                    if chat_ids:
                        
                        if not isinstance(chat_ids, list):
                            chat_ids: list = [chat_ids]
                            
                        for chat_id in chat_ids:
                            for msg in SetClient(self.auth).get_messages_updates(chat_id).get('data').get('updated_messages'):
                                if not msg.get('last_message').get('message_id') in open(self.auth+'-ids.sty', 'r').read().split('\n'):
                                    
                                    if pattern and isinstance(pattern, tuple) or isinstance(pattern, list):
                                        if search(pattern[0], msg.get('last_message').get('text')):
                                            yield [msg.get('last_message'), pattern[1]]
                                    else:
                                        yield msg.get('last_message')
                                
                                self.__appender(msg.get('last_message').get('message_id'), 'edit')
                except Exception:
                    ...

    def hand_shake(self) -> (dict):
        for message in WebSocket(self.auth).connection:
            yield message

    def handler(self, builder: object) -> (typing.Union[object,
                                                        iter, None]):
        
        '''
        ## EXAMPLE:
            
            from rb import Handler, EventBuilder, Filters

            client = Handler(...)
            
            # the funcs: `ChatsUpdates`, `MessagesUpdates`, `HandShake` # websocket
            
            client.add_event_handling(func='HandShake', events=dict(get_messages=True, get_chats=True))
            
            @client.handler
            def update(app: RubikaClient, message: EventBuilder, event):
                ...
        
        '''

        def decorator():

            methods = SetClient(self.auth, return_data_an_dict=False)
            update = Handler(self.auth)
            
            if self.handling.get('events'):
                if not isinstance(self.handling['events'], dict):
                    self.handling.update({'events': dict(get_messages=True, get_chats=True)})
            else:
                self.handling.update({'events': dict(get_messages=True, get_chats=True)})
            if not self.handling.get('func') and isinstance(self.handling.get('func'), str) and self.handling.get('func').lower() in ['handshake', 'chatsupdates', 'messagesupdates']:
                self.handling.update({'func': 'chatsupdates'})
            
            for (event) in (update.handle(method=self.handling.get('func'), **self.handling.get('events') or {})):
                event.update({'auth': self.auth, 'chat_id': self.chat_id or ''})
                message = EventBuilder(event)
                builder(methods, message, event)

        self.__appender(action='create')

        while 1:
            try:
                decorator()
            except Exception:
                pass

    def command_handler(self, func: object, *args): # to handling with custom func

        '''
        from rb import Handler, Filters, Performers

        client = Handler('session')

        def event(message):
            message.respond(message.pattern) # Filters.author - send message to user
        
        client.add_event_handling(Performers.chats_updates, event=dict(get_chats=True, get_messages=True, pattern=(client.regex('/start'), 'Hi from rubx lib.')))
        client.starting = True
        client.command_handler(event)
        '''

        def updater(*args):

            @self.handler
            def update(app, message, event):
                func(message)
        if (self.starting):
            self.__appender(action='create')
            updater()

    def on(self, event) -> (typing.Union[object, None]):
        
        '''
        ## EXAMPLE:

            from rb import Handler, NewMessage
            
            client = Handler(...)
            
            @client.on(NewMessage(client.handle, handle_name='ChatsUpdates').builder())
            def update(event):
                ...
        
        '''
        
        def decorator(func) -> (None):
            return func

        return decorator

    def add_event_handling(self, **handlers) -> (None):
        # add a handler method and params
        self.handling.update({'func': handlers.get('func') or handlers.get('method'), 'events': handlers.get('events') or handlers.get('event')})
    
    def remove_event_handling(self, func: object):
        try:
            self.handling.pop(func)
        except KeyError:
            ...

    def regex(self, word: str) -> (compile):
        return compile(word)


# to finding all attr's
class Classer(object):

    @classmethod
    def create(cls, name, __base, authorise: list = [],
               exception: bool = True, *args, **kwargs) -> object:
        
        result = None
        if authorise.__contains__(name):
            result = name

        else:
            attr = difflib.get_close_matches(name, authorise, n=1)
            
            if attr:
                return getattr(__base[0], attr[0])
            
            else:
                caller = inspect.getframeinfo(inspect.stack()[2][0])
                warn(
                    f'{caller.filename}:{caller.lineno}: do you mean'
                    f' "{name}", "{result}"? correct it')

        if result.__ne__(None) or not exception:
            if result == None:
                result = name
            # setattr(___base[0], result or name, lambda *args, **kwargs: ...)
            # return getattr(__base[0], name)
            return type(result, __base, {'__name__': result, **kwargs}) # add method to class

        raise AttributeError(f'module has no attribute ({name})')


# main class object to set all method for use.
class RubikaClient(SetClient): # TODO: add all methods
    
    __slots__ = ('SetClient', )
    
    def __init__(self, *args, **kwargs) -> None:
        
        '''
        `# rubika > rub > rb`
        
        `import rb`
        `print(help(rb))`
        
        `from rb import *`
        `with RubikaClient('session', base_logger=__name__) as client:`
            `client.proxy = {'http': '127.0.0.1:9050'}`
            `print(client * 'chat-id')`
        '''

        super().__init__(*args, **kwargs)

    def __getattr__(self, name, *args, **kwargs) -> Classer:
        
        # `note`: for: if you forget the method name
        '''
        from rb import RubikaClient
        
        with RubikaClient('session') as client:
            print(client.getChatInfo(client, 'chat-guid')) # GetChatInfo, GETchatINFO, or ...
        '''

        # for: normally
        '''
        from rb import RubikaClient
        
        with RubikaClient('session') as client:
            print(client.get_chat_info('chat-guid'))
        '''
        
        method = Classer.create(name, (SetClient, ), dir(SetClient))
        
        try:
            return method(*args, **kwargs)
        except Exception:
            return method


# To use the async for methods
# TODO: set handler with async.
class Client(RubikaClient):
    
    __slots__ = ('RubikaClient', )
    
    def __init__(self, *args, **kwargs):

        '''
        from rb import Client

        async def run(*args):
            async with Client(...) as client:
                await client.start(client.send_message, 'Hey! from rubx', 'chat-guid')
        
        Client.run(run)
        '''

        super().__init__(*args, **kwargs)
    
    async def __aenter__(self):
        return self

    async def __aexit__(self, *args, **kwargs) -> (None):
        pass

    async def start(self, method: object, *args, **kwargs) -> (dict):
        '''
        get method func to async
        '''
        return method(*args, **kwargs)
    
    @staticmethod
    def run(func: object, *args) -> (None):
        '''
        run main func to use
        
        func | func()
        -------------
        '''
        
        try:
            __import__('asyncio').run(func())
        except Exception:
            __import__('asyncio').run(func)


class BotAPI(ABC):

    __slots__ = ('self', 'base_logger', 'token',
                 'proxy', 'timeout', 'headers', 'return_data_action')

    def __str__(self) -> str:
        return dumps({'session': SQLiteSession(self.token).information()}, indent=4) if self.token else dumps({'__base_class__': 'BotAPI'})

    def __init__(
        self,
        base_logger         : str = __name__,
        token               : str = None,
        proxy               : dict = {'http': '127.0.0.1:9050'},
        timeout             : int  = 5,
        headers             : dict = None,
        return_data_action  : bool = True
        ) -> None:

        '''
        `with BotAPI(__name__, 'token') as app:`
            `app.send_message('chat-id', 'Hey!')`

        token: is api key
        proxy: to set and config proxy for requests
        headers: to set header
        return_data_action: a boolean type, literal: True, False for return keys to obejct or attrbute or dict ...
        '''

        if isinstance(base_logger, str):
            
            try:
                base_logger = logging.getLogger(base_logger)
            except Exception:
                base_logger = logging.getLogger(__name__)

        if token:

            try:
                open(f'{token}.sty', 'r')
            except FileNotFoundError:
                open(f'{token}.sty', 'w+').write(token)
                database = SQLiteSession(token)
                database.insert(None, token, None, 'https://messengerg2b1.iranlms.ir/v3/')
        else:
            
            try:
                token: str = SQLiteSession(token).information()[1]
            except Exception:
                warn('TokenWarn: please insert bot token')
        
        (self.base_logger, self.token, self.proxy,
         self.headers, self.return_data_action, self.handling) = (base_logger, token, proxy,
                                                   headers, return_data_action, {})
        (Connection.timeout, Connection.proxy,
         Connection.headers, Connection.type) = (timeout or 5, proxy,
                                                 headers, return_data_action)


    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass
    
    @property
    def get_token(self) -> str:
        return self.token


class SetAPIS(BotAPI, BotMethods): ...


class MessageHandler(object):

    def __str__(self) -> (str):
        return dumps(self.result_data, indent=4,
                     ensure_ascii=False, default=str)

    def __getattr__(self, name) -> (list):
        return self.keys_finder(keys=name)

    def __setitem__(self, key, value) -> (None):
        self.result_data[key] = value

    def __getitem__(self, key):
        return self.result_data[key]

    def __lts__(self, update: list) -> (list):
        
        for index, element in enumerate(update):
            if isinstance(element, list):
                update[index] = self.__lts__(update=element)

            elif isinstance(element, dict):
                update[index] = MakeDataKeys(update=element)

            else:
                update[index] = element
        return update

    def __init__(self, update: dict = None) -> (None):
        
        '''
        message finder
        '''
        
        self.result_data = update

    def keys_finder(self, keys: list,
                  result_data: str = None) -> (list):

        if not result_data:
            result_data = self.result_data

        if not isinstance(keys, list):
            keys = [keys]

        if isinstance(result_data, dict):
            for key in keys:
                try:
                    update = result_data[key]
                    if isinstance(update, dict):
                        update = MakeDataKeys(update=update)

                    elif isinstance(update, list):
                        update = self.__lts__(update=update)

                    return update

                except KeyError:
                    pass
            result_data = result_data.values()

        for value in result_data:
            if isinstance(value, (dict, list)):
                try:
                    return self.keys_finder(keys=keys, result_data=value)

                except AttributeError:
                    pass
        else:
            raise KeyError('This value is not found')

    @property
    def raw_text(self):
        return self.keys_finder(['text'])

    @property
    def message_id(self):
        return self.keys_finder(['message_id'])

    @property
    def reply_to_message_id(self):
        return self.keys_finder(['reply_to_message_id'])

    @property
    def chat_id(self):
        return self.keys_finder(['chat_id'])

    def reply(self, text: str) -> typing.Union[str, dict]:
        return SetAPIS(token=self.token).send_message(self.chat_id, text, reply_to_message_id=self.message_id)


# the main class to usage rubika bot api's
class BotAPI(SetAPIS):


    def __init__(self, *args, **kwargs) -> None:
        '''
        bot methods api client
        '''
        
        super().__init__(*args, **kwargs)


    def __getattr__(self, name, *args, **kwargs) -> typing.TypeVar:
        
        # `note`: for: if you forget the method name
        '''
        from rb import BotAPI

        with BotAPI(__name__, 'token') as app:
            print(client.sendMessage(app, 'chat-guid', 'Hey')) # SendMessage, SENDmessage, sendMESSAGE, or ...
        '''

        # for: normally
        '''
        from rb import BotAPI
        
        with BotAPI(__name__, 'token') as app:
            print(app.send_message('chat-guid', 'Hey'))
        '''
        
        method = Classer.create(name, (SetAPIS, ), dir(SetAPIS))
        
        try:
            return method(*args, **kwargs)
        except Exception:
            return method


    def add_event_handling(
        pattern: typing.Union[typing.Tuple[str],
                              typing.List[str]],
        *args,
        **kwargs
        ) -> None:
        
        '''
        `self.add_event_handling(pattern=(r'\w{1}start', 'Hey!'))`
        '''
    
        if not isinstance(pattern, (tuple, list)):
            raise ValueError('The \'pattern\' is invalid, pattern is not a tuple or list')
        else:
            if pattern.__len__() < 2:
                raise IndexError('The \'pattern\' is invalid, len pattern is not 2')
            
        self.handling.update({'pattern': pattern})


    def handler(self, builder: object, __limit: int = 1) -> None:

        '''
        with BotAPI(__name__, 'token') as app:
            app.add_event_handling(pattern=('\w{1}start', 'Hey!'))
            
            @app.handler
            def update(methods, update, event):
                ...
        
        '''

        def decorator(*args, **kwargs) -> typing.Any:
            
            pattern, methods = None, SetAPIS(token=self.token)
            
            if self.handling.get('pattern'):
                pattern: typing.Union[list, tuple] = self.handling['pattern']
            
            while 1:
                
                for event in self.get_updates(__limit).get('updates'):
                    event.update({'pattern': pattern[1], 'token': self.token})
                    
                    if pattern:
                            update = MessageHandler(event)
                            if re.compile(pattern[0]).search(update.raw_text):
                                builder(methods, update, event)
                    else:
                        update = MessageHandler(event)
                        builder(methods, update, event)

        while 1:
            try:
                decorator()
            except Exception:
                pass


class StartClient(RubikaClient): ...


__all__ = ['Client', 'RubikaClient',
           'RubinoClient', 'WebSocket',
           'Handler', 'NewMessage',
           'EventBuilder', 'Filters',
           'Performers', 'Method', 'Attrs']

__version__ = Version.__version__


#!/bin/python

import pip, warnings, typing


class PyPi(object):

    @classmethod
    def __init__(cls) -> None:
        pass

    def __installer(cls, packages: typing.Union[list, tuple] = [],
                    commands: list = None, action: str = 'install'):

        if not commands:
            for module in packages:
                try:
                    __import__(module if module != 'pycryptodome' else 'Crypto')
                    
                except Exception:
                    pip.main([action or 'install', '--user', module])
        else:
            
            if not isinstance(commands, list):
                commands: list = [commands]
            if len(commands) is not len(packages):
                warnings.warn('len commands is not len packages')
            
            else:
                for module, cmd in zip(packages, commands):
                    
                    try:
                        __import__(module)
                    except Exception:
                        pip.main([cmd or 'install', '--user', module])


    def installation(cls, modules: typing.Union[list, tuple],
                     commands: typing.Union[list, tuple] = None, action: str = 'install') -> None:
        
        '''
        `modules`: type is list[str] or tuple[str]
        
        `PyPi().installation(['tinytag', 'webbrowser', 'pycryptodome'])`
        '''

        cls.__installer(modules, commands, action)